/**
 ** sahandCntrlTaskCodels.cc
 **
 ** Codels used by the control task sahandCntrlTask
 **
 ** Author:
 ** Date:
 **
 **/

#include <portLib.h>
#include <h2mathLib.h>

#include "server/sahandHeader.h"
#include "sahandConst.h"
#include "sahandPrivate.h"

static STATUS
sahandControlPort(int port, int *report)
{
	
	if (port < SAH_PORT_1 || port > SAH_PORT_2) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (!SAHAND_PORT_ACTIVE(port)) {
		*report = S_sahand_INACTIVE_PORT;
		return ERROR;
	}
	return OK;
}

/*------------------------------------------------------------------------
 * sahandSetControllerControl  -  control codel of EXEC request SetController
 *
 * Description: check the validity of the controller type
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_DRIVER_ERROR
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandSetControllerControl(SAHAND_THIS_CONTROLLER_TYPE *sah_controller, 
    int *report)
{
	int port = sah_controller->stct_port;
	SAHAND_CONTROLLER_TYPE controller = sah_controller->stct_controller;
	
	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	if (controller < SAH_CONTROLLER_RESET ||
	    controller > SAH_CONTROLLER_IMPEDANCE) {
		*report =  S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	SDI_F->sah_params[port].sap_controller = controller;
	SDI_F->sah_command[port] |= SAH_COMMAND_SET_CONTROLLER;
	return OK;
}

/*------------------------------------------------------------------------
 * sahandMoveFingerControl  -  control codel of EXEC request MoveFinger
 *
 * Description: check bounds of finger move commands
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_DRIVER_ERROR
 *
 * Returns:    OK or ERROR
 */


STATUS
sahandMoveFingerControl(SAHAND_THIS_FINGER_REF *sah_finger_ref, int *report)
{
	double *angle, *vel;
	SAHAND_FINGER_REF *ref;
	int port = sah_finger_ref->stfr_port;
	
	ref = &SDI_F->sah_ref[port].sr_fingers[0];
	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	angle = &sah_finger_ref->stfr_finger.sfr_angle[0];
	vel = &sah_finger_ref->stfr_finger.sfr_vel[0];
	/* Check finger Id  */
	if (sah_finger_ref->stfr_id < 0 ||
	    sah_finger_ref->stfr_id >= SAH_MAX_FINGERS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	/* Check values */
	if (angle[0] < SAH_MIN_DISTAL_ANGLE  ||
	    angle[0] > SAH_MAX_DISTAL_ANGLE) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (angle[1] < SAH_MIN_PROXIMAL_ANGLE ||
	    angle[1] > SAH_MAX_PROXIMAL_ANGLE) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (angle[2] < SAH_MIN_BASE_ANGLE ||
	    angle[2] > SAH_MAX_BASE_ANGLE) {
		*report =  S_sahand_BAD_PARAMETERS;
		return ERROR;
	}

	if (vel[0] < 0.0 ||
	    vel[0] > SAH_MAX_DISTAL_VEL) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (vel[1] < 0.0 ||
	    vel[1] > SAH_MAX_PROXIMAL_VEL) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (vel[2] < 0.0 ||
	    vel[2] > SAH_MAX_BASE_VEL) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}

	/* copy data to finger struct */
	memcpy(&ref[sah_finger_ref->stfr_id],
	    &sah_finger_ref->stfr_finger,
	    sizeof(SAHAND_FINGER_REF));

	SDI_F->sah_command[port] |= SAH_COMMAND_MOVE_HAND;
	return OK;
}

/*------------------------------------------------------------------------
 * sahandSetStiffnessControl  -  control codel of EXEC request SetStiffness
 *
 * Description: checks bounds of the stiffness factor command.
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_DRIVER_ERROR
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandSetStiffnessControl(SAHAND_THIS_FINGER_STIFFNESS *sah_finger_stiffness,
    int *report)
{
	double *value = &sah_finger_stiffness->stfs_finger.ssf_value[0];
	int i, finger = sah_finger_stiffness->stfs_id,
	    port = sah_finger_stiffness->stfs_port;

	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	/* Check finger Id  */
	if (finger < 0 || finger >= SAH_MAX_FINGERS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	/* Check values */
	if (value[0] < 0.0 ||
	    value[0] > SAH_MAX_DISTAL_STIFFNESS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (value[1] < 0.0 ||
	    value[1] > SAH_MAX_PROXIMAL_STIFFNESS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	if (value[2] < 0.0 ||
	    value[2] > SAH_MAX_BASE_STIFFNESS) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	/* copy data */
	for (i = 0; i < 3; i++) 
		SDI_F->sah_params[port].sap_stiffness[finger].ssf_value[i] = 
		    value[i];
	SDI_F->sah_command[port] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
	return OK;
}

/*------------------------------------------------------------------------
 * sahandRotateThumbControl  -  control codel of EXEC request RotateThumb
 *
 * Description: check bounds of the thumb angle command
 *              and register the command to be executed
 * Report: OK
 *                 S_sahand_BAD_PARAMETERS
 *                 S_sahand_DRIVER_ERROR
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandRotateThumbControl(SAHAND_THIS_THUMB_ANGLE *sah_thumb_angle, int *report)
{
	int port = sah_thumb_angle->stta_port;
	double angle = sah_thumb_angle->stta_angle;

	if (sahandControlPort(port, report) == ERROR)
		return ERROR;
	if (angle < SAH_MIN_THUMB_ORIENTATION ||
	    angle > SAH_MAX_THUMB_ORIENTATION) {
		*report = S_sahand_BAD_PARAMETERS;
		return ERROR;
	}
	SDI_F->sah_ref[port].sr_thumb_orientation = angle;
	SDI_F->sah_command[port] |= SAH_COMMAND_SET_THUMB_ANGLE;
	return OK;
}
