/**
 ** sahandMainCodels.cc
 **
 ** Codels called by execution task sahandMain
 **
 ** Author:
 ** Date:
 **
 **/

#include <math.h>

#include <portLib.h>
#include <h2mathLib.h>

#include <SAHandCtrlApi.h>

#include "server/sahandHeader.h"
#include "sahandConst.h"
#include "sahandPrivate.h"

static CSAHandCtrlApi sahapi;		// Schunk API object

/*----------------------------------------------------------------------*/
/* Convert a Genom port number/index to a Schunk API port */
static int
sahandPort(int p)
{
	switch (p) {
	case SAH_PORT_1: return PORT_1;
	case SAH_PORT_2: return PORT_2;
	default: return -1;
	}
}
/*----------------------------------------------------------------------*/
static STATUS
sahandOneInit(int hand, int *report)
{
	float pos[3], vel[3], stiffness[3];
	SAHAND_FINGER_REF *f;
	int i, j, angle;
	int port = sahandPort(hand);
	short res;

	/* Default speeds */
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		f = &(SDI_F->sah_ref[hand].sr_fingers[i]);
		for (j = 0; j < 3; j++)
			f->sfr_vel[j] = DEG_TO_RAD(100.0);
	}
	/* Controller defaults to IMPEDANCE */
	res = sahapi.SetController(port, CONTROLLER_IMPEDANCE);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_params[hand].sap_controller = SAH_CONTROLLER_IMPEDANCE;

	/* Release brakes */
	res = sahapi.SetEmergencyStop(port, RUN);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_brake = SAH_RUN;
	res = sahapi.SetThumbBrake(port, RUN);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_thumb.sts_brake = SAH_RUN;

	/* Set all joints references to current pos, and 0 stiffness */
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		f = &(SDI_F->sah_ref[hand].sr_fingers[i]);
		for (j = 0; j < 3; j++) {
			vel[j] = RAD_TO_DEG(f->sfr_vel[j]);
			stiffness[j] = 0.0;
		}
		sahapi.SetFingerEnable(port, i, TRUE);
		sahapi.GetJointAngle(port, i, pos);
		/* Set reference to current position */
		for (j = 0; j < 3; j++)
			f->sfr_angle[j] = DEG_TO_RAD(pos[j]);
		sahapi.MoveFinger(port, i, pos, vel);
		sahapi.SetStiffnessFactor(port, i, stiffness);
	}
	
	/* Update thumb reference  */
	sahapi.GetThumbAngle(port, &angle);
	SDI_F->sah_ref[hand].sr_thumb_orientation = DEG_TO_RAD((double)angle);
	return OK;
}
/*------------------------------------------------------------------------
 *
 * sahandInit  --  Initialization codel (fIDS, ...)
 *
 * Description: connects to the driver, sets all parameters to a known state.
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandInit(int *report)
{
	int i, config, port;
	short res;

	res = sahapi.SAHandInit();
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	for (i = 0; i < 2; i++) {
		port = sahandPort(i);
		/* Is it a left or right hand ? */
		res = sahapi.GetHandConfig(port, &config);
		if (res <= 0) {
			/* error */
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		} else {
			switch(config) {
			case 0:
				SDI_F->sah_state[i].ss_config = 
				    SAH_CONFIG_NOT_PRESENT;
				break;
			case HAND_CONFIG_LEFT:
				SDI_F->sah_state[i].ss_config = 
				    SAH_CONFIG_LEFT;
				break;
			case HAND_CONFIG_RIGHT:
				SDI_F->sah_state[i].ss_config = 
				    SAH_CONFIG_RIGHT;
				break;
			default:
				*report = S_sahand_DRIVER_ERROR;
				return ERROR;
			} /* case */
			/* init this hand */
			if (sahandOneInit(i, report) == ERROR)
				return ERROR;
		}
	} /* for */
	return OK;
}

/*------------------------------------------------------------------------
 *
 * sahandEnd  --  Termination codel
 *
 * Description: puts brakes on, disconnect from the driver.
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandEnd(void)
{
	int i, port;
	
	for (i = 0; i < 2; i++) {
		if (!SAHAND_PORT_ACTIVE(i)) 
			continue;
		port = sahandPort(i);
		sahapi.SetEmergencyStop(port, TRUE);
		sahapi.SetThumbBrake(port, TRUE);
	}
	return OK;
}

/*----------------------------------------------------------------------*/
/**
 ** Helper functions for the permanent activity
 **
 ** Wrapping the API provided by Schunk
 **/

static short
sahandSetController(int hand)
{
	int controller = SDI_F->sah_params[hand].sap_controller,
	    port = sahandPort(hand);

	return sahapi.SetController(port, controller);
}

static short
sahandSetThumbAngle(int hand)
{
	int angle = (int)RAD_TO_DEG(SDI_F->sah_ref[hand].sr_thumb_orientation);
	int port = sahandPort(hand);

	return sahapi.SetThumbAngle(port, angle);
}

static short
sahandSetThumbBrake(int hand)
{
	return 0;
}

static short
sahandMoveHand(int hand)
{
	SAH_DESIRED desired;
	SAHAND_FINGER_REF *finger;
	int i, j, port = sahandPort(hand);

	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_THUMB];
	/* Convert parameters for the Schunk API */
	for (i = 0; i < 3; i++) {
		desired.Thumb.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.Thumb.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
	}
	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_FIRST];
	for (i = 0; i < 3; i++) {
		desired.FirstFinger.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.FirstFinger.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
	}
	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_MIDDLE];
	for (i = 0; i < 3; i++) {
		desired.MiddleFinger.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.MiddleFinger.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
	}
	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_RING];
	for (i = 0; i < 3; i++) {
		desired.RingFinger.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.RingFinger.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
	}
	return sahapi.MoveHand(port, &desired);
}

static short
sahandSetStiffness(int hand)
{
	double *ref;
	float stiffness[3];
	int i, j;
	int port = sahandPort(hand);
	short res;
	
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		ref = &SDI_F->sah_params[hand].sap_stiffness[i].ssf_value[0];
		for (j = 0; j < 3; j++)
			stiffness[j] = (float)ref[j];
		res = sahapi.SetStiffnessFactor(port, i, stiffness);
		if (res <= 0) 
			break;
	}
	return res;
}

/*----------------------------------------------------------------------*/

/**
 ** Send pending action commands to the driver
 **/
static STATUS
sahandAction(int hand, int *report)
{
	int command = SDI_F->sah_command[hand];
	short res;

	/* Pending actions are set in the control task */
	if (command & SAH_COMMAND_SET_CONTROLLER)
		res = sahandSetController(hand);
	if (command & SAH_COMMAND_SET_THUMB_ANGLE)
		res = sahandSetThumbAngle(hand);
	if (command & SAH_COMMAND_SET_THUMB_BRAKE)
		res = sahandSetThumbBrake(hand);
	if (command & SAH_COMMAND_MOVE_HAND)
		res = sahandMoveHand(hand);
	if (command & SAH_COMMAND_SET_STIFFNESS_FACTOR)
		res = sahandSetStiffness(hand);
	/* Store last result */
	SDI_F->sah_state[hand].ss_error = res;
	/* Clear pending actions */
	SDI_F->sah_command[hand] = 0; 		
	return OK;
}

/*----------------------------------------------------------------------*/
/**
 ** Read sensor values
 **/
static STATUS
sahandSensors(int hand, int *report)
{
	float temp, pos[3];
	SAHAND_FINGER_STATE *f;
	int i, j, tpos, port = sahandPort(hand);
	short res;

	/* Fingers */
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		f = &(SDI_F->sah_state[hand].ss_finger[i]);
		/* Position */
		res = sahapi.GetJointAngle(port, i, pos);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		for (j = 0; j < 3; j++)
			f->sfs_pos[j] = DEG_TO_RAD(pos[j]);
		/* speed */
		res = sahapi.GetJointSpeed(port, i, pos);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		for (j = 0; j < 3; j++)
			f->sfs_vel[j] = DEG_TO_RAD(pos[j]);
		/* torque */
		res = sahapi.GetJointTorque(port, i, pos);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		for (j = 0; j < 3; j++)
			f->sfs_torque[j] = pos[j];
		/* Finger temperature */
		res = sahapi.GetFingerTemp(port, i, &temp);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		f->sfs_finger_temp = (double)temp;
		/* Joint limits */
		res = sahapi.GetFingerLimitsStatus(port, i, f->sfs_limits);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		
	}
	/* Thumb position */
	res = sahapi.GetThumbAngle(port, &tpos);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_thumb.sts_angle = DEG_TO_RAD((double)tpos);

	/* Palm temperature */
	res = sahapi.GetPalmTemp(port, &temp);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_palm_temp = temp;
	return OK;
}
/*------------------------------------------------------------------------
 *
 * sahandMainPerm  --  codel of permanent activity 
 *                     (called BEFORE other activities)
 *
 * Description: read sensors and execute actions
 *
 * Reports:      OK
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandMainPerm(int *report)
{
	int hand;

	for (hand = 0; hand < 2; hand++) {
		if (!SAHAND_PORT_ACTIVE(hand)) 
			continue;

		/* Sensors */
		if (sahandSensors(hand, report) == ERROR)
			return ERROR;
	
		/* Actions */
		if (SDI_F->sah_command[hand] != 0) 
			if (sahandAction(hand, report) == ERROR)
				return ERROR;
	}
	return OK;
}

/*------------------------------------------------------------------------
 * ClearTorqueOffset
 *
 * Description: clear existing offset on sensors for one finger
 *
 * Reports:      OK
 *              S_sahand_DRIVER_ERROR
 *              S_sahand_INACTIVE_PORT
 */

/* sahandClearTorqueOffset  -  codel EXEC of ClearTorqueOffset
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandClearTorqueOffset(SAHAND_THIS_FINGER *sah_finger_id, int *report)
{
	short res;

	res = sahapi.ClearTorqueSensorOffset(
	    sahandPort(sah_finger_id->stf_port), sah_finger_id->stf_id);
	SDI_F->sah_state[sah_finger_id->stf_port].ss_error = res;
	
	return ETHER;
}

/*------------------------------------------------------------------------
 * Brakes
 *
 * Description: Set brakes
 *
 * Reports:      OK
 *              S_sahand_DRIVER_ERROR
 */

/* sahandStop  -  codel EXEC of Stop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandBrake(SAHAND_THIS_HAND_BRAKE *sah_brake, int *report)
{
	short res;

	res = sahapi.SetEmergencyStop(sah_brake->sthb_port, 
	    sah_brake->sthb_brake);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return FAIL;
	}
	return ETHER;
}
